/* This Applet simulates the behaviour of an N-station*/
/* request service flow system. Input parameters are:*/
/* name=BACKGROUND_COLOR    value=RGB value*/
/* name=INPUT_RATE          value=request arrival rate*/
/* name=N_QUEUES            value=number of service stations*/
/* name=QUEUE_COLOR_n       value=RGB value of color for requests at each */
/*                                station*/
/* name=QUEUE_SERVICETIME_n value=service time for each station*/
/* name=QUEUE_STICKINESS_n  value=fraction of system load consumed by idle */
/*                                jobs at each station*/
/* name=QUEUE_TIMEOUT_n     value=timeout for job execution at each station*/
/* This was written a bit in a hurry, and would need lots of comments. */
/* francesco.prelz@mi.infn.it 20030220*/
 
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.applet.Applet;
import java.io.IOException;

public class StickyQueue extends Applet implements MouseListener
 {
  ClockThread secthread;
  int sofx, sofy;
  int n_queues;
  double in_rate;
  QueueStack queuestack[];
  TextField in_timeout[];
  TextField in_stickiness[];
  TextField in_serviceticks[];
  TextField in_ratef;
  double q_stickiness[];
  double q_timeout[];
  double q_serviceticks[];
  ThermoGauge loadg;
  ThermoGauge success;
  Button resetbut;
  Label stimel,soutrl;
  int stime;

  void init_parameters()
   {
    String temp;
    int i;

    if (n_queues > 0)
     {
      for (i=0; i<n_queues; i++)
       {
        temp = getParameter("QUEUE_TIMEOUT_"+(i+1));
        if (temp != null)
         {
          try
           {
            q_timeout[i] = Double.valueOf(temp).doubleValue();
           } catch (NumberFormatException e) {
            q_timeout[i] = 0;
           }
         }
        else q_timeout[i] = 0;
        temp = getParameter("QUEUE_STICKINESS_"+(i+1));
        if (temp != null)
         {
          try
           {
            q_stickiness[i] = Double.valueOf(temp).doubleValue();
           } catch (NumberFormatException e) {
            q_stickiness[i] = 0;
           }
         }
        else q_stickiness[i] = 0;
        temp = getParameter("QUEUE_SERVICETIME_"+(i+1));
        if (temp != null)
         {
          try
           {
            q_serviceticks[i] = Double.valueOf(temp).doubleValue();
           } catch (NumberFormatException e) {
            q_serviceticks[i] = 0;
           }
         }
        else q_serviceticks[i] = 0;
       }
     }  

    temp = getParameter("INPUT_RATE");
    if (temp != null)
     {
      try
       {
        in_rate = Double.valueOf(temp).doubleValue();
       } catch (NumberFormatException e) {
        in_rate = 0;
       }
     }
   }

  public void init() 
   {
    GridBagLayout gridBag = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();

    Font general_font = new Font("dialog", Font.PLAIN, 16);

    setLayout(gridBag);

    String temp = getParameter("X_OFFSET");
    if (temp != null)
     {
      try
       {
        sofx = Integer.parseInt(temp);
       } catch (NumberFormatException e) {
       }
     }

    temp = getParameter("Y_OFFSET");
    if (temp != null)
     {
      try
       {
        sofy = Integer.parseInt(temp);
       } catch (NumberFormatException e) {
       }
     }

    temp = getParameter("BACKGROUND_COLOR");
    if (temp != null)
     {
      try
       {
        int bgcolor = Integer.parseInt(temp);
        this.setBackground(new Color(bgcolor));
       } catch (NumberFormatException e) {
       }
     }

    temp = getParameter("N_QUEUES");
    if (temp != null)
     {
      try
       {
        n_queues = Integer.parseInt(temp);
       } catch (NumberFormatException e) {
        n_queues = 0;
       }
     }
    else n_queues = 0;

    if (n_queues > 0)
     {
      queuestack = new QueueStack[n_queues];
      in_timeout = new TextField[n_queues];
      in_stickiness = new TextField[n_queues];
      in_serviceticks = new TextField[n_queues];
      q_stickiness = new double[n_queues];
      q_timeout = new double[n_queues];
      q_serviceticks = new double[n_queues];
      int i;
      int color=255;
      for (i=0; i<n_queues; i++)
       {
        temp = getParameter("QUEUE_COLOR_"+(i+1));
        if (temp != null)
         {
          try
           {
            color = Integer.parseInt(temp);
           } catch (NumberFormatException e) {
            color = 255;
           }
         }
        queuestack[i] = new QueueStack(new Color(color));
       }
     }

    init_parameters();

    int i;
    ParameterDispatch pdisp;
    for (i=0; i<n_queues; i++)
     {
      in_timeout[i]      = new TextField(Double.toString(q_timeout[i]),10); 
      in_timeout[i].setFont(general_font);
      in_timeout[i].setBackground(Color.white);
      in_stickiness[i]   = new TextField(Double.toString(q_stickiness[i]),10); 
      in_stickiness[i].setFont(general_font);
      in_stickiness[i].setBackground(Color.white);
      in_serviceticks[i] = new TextField(Double.toString(q_serviceticks[i]),10); 
      in_serviceticks[i].setFont(general_font);
      in_serviceticks[i].setBackground(Color.white);
      pdisp = new ParameterDispatch(this, i, in_timeout[i], in_stickiness[i], in_serviceticks[i]);
      in_timeout[i].addTextListener(pdisp);
      in_stickiness[i].addTextListener(pdisp);
      in_serviceticks[i].addTextListener(pdisp);
     }

    c.fill = GridBagConstraints.BOTH;
    c.weightx = 1.0;
    c.weighty = 0.0;
    c.gridwidth = 1;
    c.gridheight = 1;
    c.gridx = 0;
    c.gridy = GridBagConstraints.RELATIVE;
    Label label1 = new Label("System load",Label.CENTER);
    label1.setFont(general_font);
    gridBag.setConstraints(label1,c);
    add(label1);
    loadg = new ThermoGauge(100,Color.red);
    loadg.setSecondaryColor(new Color(255,165,0));
    gridBag.setConstraints(loadg,c);
    add(loadg);
    stime = 0;
    stimel = new Label("",Label.CENTER);
    stimel.setFont(general_font);
    gridBag.setConstraints(stimel,c);
    add(stimel);
    Label labeli = new Label("Arrival rate:");
    labeli.setFont(general_font);
    gridBag.setConstraints(labeli,c);
    add(labeli);
    in_ratef     = new TextField(Double.toString(in_rate),10); 
    in_ratef.setFont(general_font);
    in_ratef.setBackground(Color.white);
    pdisp = new ParameterDispatch(this);
    in_ratef.addTextListener(pdisp);
    gridBag.setConstraints(in_ratef,c);
    add(in_ratef);
    c.gridx = 1;
    c.gridy = GridBagConstraints.RELATIVE;
    Label label6 = new Label("Success rate",Label.CENTER);
    label6.setFont(general_font);
    gridBag.setConstraints(label6,c);
    add(label6);
    success = new ThermoGauge(100,new Color(0,150,0));
    gridBag.setConstraints(success,c);
    add(success);
    Label label7 = new Label("Out rate:");
    label7.setFont(general_font);
    gridBag.setConstraints(label7,c);
    add(label7);
    soutrl = new Label("",Label.CENTER);
    soutrl.setFont(general_font);
    gridBag.setConstraints(soutrl,c);
    add(soutrl);
    resetbut = new Button("Reset");
    resetbut.setFont(general_font);
    resetbut.addMouseListener(this);
    gridBag.setConstraints(resetbut,c);
    add(resetbut);

    c.gridheight = 5;
    c.weighty = 1.0;
    c.gridx = GridBagConstraints.RELATIVE;
    c.gridy = GridBagConstraints.RELATIVE;
    for (i=0;i<n_queues;i++)
     {
      if (i == (n_queues-1) )
        c.gridwidth = GridBagConstraints.REMAINDER; //end row
      gridBag.setConstraints(queuestack[i],c);
      add(queuestack[i]);
     }
    c.gridheight = 1;
    c.fill = GridBagConstraints.BOTH;
    c.weightx = 1.0;
    c.weighty = 0.0;
    c.gridwidth = 2;
    Label label2 = new Label("Queue timeout");
    label2.setFont(general_font);
    gridBag.setConstraints(label2,c);
    add(label2);
    c.gridwidth = 1;
    for (i=0;i<n_queues;i++)
     {
      if (i == (n_queues-1) )
        c.gridwidth = GridBagConstraints.REMAINDER; //end row
      gridBag.setConstraints(in_timeout[i],c);
      add(in_timeout[i]);
     }
    c.fill = GridBagConstraints.BOTH;
    c.weightx = 1.0;
    c.weighty = 0.0;
    c.gridwidth = 2;
    Label label3 = new Label("Queue \"stickiness\"");
    label3.setFont(general_font);
    gridBag.setConstraints(label3,c);
    add(label3);
    c.gridwidth = 1;
    for (i=0;i<n_queues;i++)
     {
      if (i == (n_queues-1) )
        c.gridwidth = GridBagConstraints.REMAINDER; //end row
      gridBag.setConstraints(in_stickiness[i],c);
      add(in_stickiness[i]);
     }
    c.fill = GridBagConstraints.BOTH;
    c.weightx = 1.0;
    c.weighty = 0.0;
    c.gridwidth = 2;
    Label label4 = new Label("Queue service time");
    label4.setFont(general_font);
    gridBag.setConstraints(label4,c);
    add(label4);
    c.gridwidth = 1;
    for (i=0;i<n_queues;i++)
     {
      if (i == (n_queues-1) )
        c.gridwidth = GridBagConstraints.REMAINDER; //end row
      gridBag.setConstraints(in_serviceticks[i],c);
      add(in_serviceticks[i]);
     }

    secthread = new ClockThread(this);
    secthread.start();
    repaint();
    validate();
   }

  public void mouseClicked(MouseEvent e) {}
  public void mousePressed(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  public void mouseReleased(MouseEvent e)
   {
    if (e.getSource() == resetbut)
     {
      init_parameters();
      loadg.setSecondaryFraction(0);
      loadg.set(0);
      success.set(0);

      init_parameters();

      in_ratef.setText(Double.toString(in_rate)); 

      int n;
      for (n=0; n<n_queues; n++)
       {
        queuestack[n].set(0);
        queuestack[n].busy=false;
        queuestack[n].n_dropped_jobs=0;
        queuestack[n].repaint();
        in_timeout[n].setText(Double.toString(q_timeout[n])); 
        in_stickiness[n].setText(Double.toString(q_stickiness[n])); 
        in_serviceticks[n].setText(Double.toString(q_serviceticks[n])); 
       }
      secthread.in_total = secthread.out_total = 0;
      stime = 0;
     }
   }

 }

class ThermoGauge extends Canvas 
 {
  int max;
  int value;
  Color gcolor;
  Color scolor;
  int drawn_value;
  int cx,miny,maxy,br,cw,bord;
  double sec_fraction,drawn_sec_fraction;
  Dimension old_dim;

  public ThermoGauge (int imax, Color c) 
   {
    max = imax;
    drawn_value = value = 0; 
    gcolor = c;
    old_dim = null;
    scolor = Color.yellow;
    drawn_sec_fraction = sec_fraction = 0;
    Dimension d = new Dimension();
    d.width = 100;
    d.height = 250;
    setSize(d);
   }

  public void setSecondaryColor (Color c)
   {
    scolor = c;
   }

  public void setSecondaryFraction (double f)
   {
    sec_fraction = f;
   }

  public void set (int new_value)
   {
    value = new_value;
    if (value < 0) value = 0;
    if (value > max) value = max;
    repaint();
   }

  public int get ()
   {
    return value;
   }

  public int add (int howmuch)
   {
    int new_value = value += howmuch;
    set(new_value);
    return value;
   }

  public void paint(Graphics g) 
   {
    Dimension d = getSize();
    cx = d.width/2;
    cw = d.width/18;  
    br = cw*3;
    miny = d.height/10;
    maxy = d.height - miny - br*2;
    bord = cw/3;

    int valy = maxy - (int)((double)(maxy - miny)/(double)max * (double)value);
    
    g.setColor(gcolor);
    g.fillRect(cx-cw,valy,cw*2,maxy-valy);
    if (sec_fraction > 0)
     {
      g.setColor(scolor);
      int sech = (int)((double)(maxy - valy)*sec_fraction);
      g.fillRect(cx-cw,valy,cw*2,sech);
      g.setColor(gcolor);
     }
    g.fillOval(cx-br,maxy-cw,br*2,br*2);
    
    g.setColor(Color.black);
    double cutarc = Math.asin((double)(cw+bord)/(double)br);
    int dby = maxy-cw-bord+br-(int)((double)br*Math.cos(cutarc));
    int dgcutarc = (int)((double)180/Math.PI * (double)cutarc);
    g.drawArc(cx-br-bord,maxy-cw-bord,br*2+bord*2,br*2+bord*2,
              90+dgcutarc, 360-2*dgcutarc);
    g.drawLine(cx-cw-bord,dby,cx-cw-bord,miny-bord);
    g.drawLine(cx-cw-bord,miny-bord,cx+cw+bord,miny-bord);
    g.drawLine(cx+cw+bord,miny-bord,cx+cw+bord,dby);

    old_dim = d;
    drawn_value = value; 
    drawn_sec_fraction = sec_fraction;

    g.setFont(new Font("dialog", Font.BOLD, (cx-cw)*18/120));
    int labval,labvaly;
    int labvalstep = max/5;
    String label;
    for (labval = max; labval > max/10; labval -= labvalstep)
     {
      labvaly = maxy - (int)((double)(maxy - miny)/(double)max * (double)labval);
      label = new String(new Integer(labval).toString());
      g.drawString(label,cx+cw+2*bord,labvaly);
     }

   }

  public void update(Graphics g) 
   {
    Dimension d = getSize();
    if (old_dim != null)
     {
      if ((d.width != old_dim.width) || (d.height != old_dim.height))
       {
        g.clearRect(0,0,d.width,d.height);
        paint(g);
        return;
       }
     }
    else return;

    int valy = maxy - (int)((double)(maxy - miny)/(double)max * (double)value);
    int dvaly = maxy - (int)((double)(maxy - miny)/(double)max * (double)drawn_value);
    
    if (value > drawn_value)
     {
      if (sec_fraction > 0)
       {
        g.setColor(gcolor);
        g.fillRect(cx-cw,valy,cw*2,maxy-valy);
        int sech = (int)((double)(maxy - valy)*sec_fraction);
        g.setColor(scolor);
        g.fillRect(cx-cw,valy,cw*2,sech);
        g.setColor(gcolor);
        g.fillOval(cx-br,maxy-cw,br*2,br*2);
       }
      else
       {
        g.setColor(gcolor);
        g.fillRect(cx-cw,valy,cw*2,dvaly-valy+1);
       }
     }
    else if (value < drawn_value)
     {
      g.clearRect(cx-cw,dvaly,cw*2,valy-dvaly);
      g.setColor(gcolor);
      if (sec_fraction > 0)
       {
        g.fillRect(cx-cw,valy,cw*2,maxy-valy);
        int sech = (int)((double)(maxy - valy)*sec_fraction);
        g.setColor(scolor);
        g.fillRect(cx-cw,valy,cw*2,sech);
        g.setColor(gcolor);
       }
      g.fillOval(cx-br,maxy-cw,br*2,br*2);
     }
    else if (sec_fraction < drawn_sec_fraction)
     {
      g.setColor(gcolor);
      g.fillRect(cx-cw,valy,cw*2,maxy-valy);
      int sech = (int)((double)(maxy - valy)*sec_fraction);
      g.setColor(scolor);
      g.fillRect(cx-cw,valy,cw*2,sech);
      g.setColor(gcolor);
      g.fillOval(cx-br,maxy-cw,br*2,br*2);
     }
    else if (sec_fraction > drawn_sec_fraction)
     {
      int sech = (int)((double)(maxy - valy)*sec_fraction);
      g.setColor(scolor);
      g.fillRect(cx-cw,valy,cw*2,sech);
      g.setColor(gcolor);
      g.fillOval(cx-br,maxy-cw,br*2,br*2);
     }
    drawn_value = value; 
    drawn_sec_fraction = sec_fraction;
   }
 }

class QueueStack extends Canvas 
 {
  int n_items, n_drawn_items;
  int n_dropped_jobs;
  int debris_histo[],hisw;
  Dimension old_dim;
  Color qcolor;
  double ticks_left;
  int start_time;
  boolean busy;
  int ih, cx, cw, sy, fh, bb, bh, bw, by;

  public QueueStack (Color c) 
   {
    qcolor = c;
    n_items = n_drawn_items = 0;
    ticks_left = 0;
    start_time = 0;
    n_dropped_jobs = 0;
    old_dim = null;
    busy = false;
   }

  public int inc () 
   {
    n_items++;
    repaint();
    return n_items;
   }

  public int dec () 
   {
    if (n_items>0)
     {
      n_items--;
      repaint();
     }
    return n_items;
   }

  public int add (int n) 
   {
    n_items+=n;
    repaint();
    return n_items;
   }

  public int set (int n) 
   {
    n_items=n;
    repaint();
    return n_items;
   }

  public int sub (int n) 
   {
    if (n_items>=n)
     {
      n_items-=n;
      repaint();
     }
    else if (n_items > 0)
     {
      n_items=0;
      repaint();
     }
    return n_items;
   }

  void init_geometry(Dimension d)
   {
    fh = d.height/5;
    bb = fh/10;
    bh = bb*6;
    ih = (d.height-bh)/40; // Height of each item
    cx = d.width/5;
    cw = d.width - 2*cx;
    sy = d.height - 2*ih - 2*fh;
    bw = d.width - 2*bb;
    by = d.height - fh - bb - bh;
   }

  public void dropjob()
   {
    Graphics g = getGraphics();
    Dimension d=getSize();
    int cy;
    
    busy = false;
    for (cy=d.height-fh-bh; cy<d.height-ih*2; cy+=ih/2)
     {
      g.setColor(qcolor);
      g.fillOval(cx,cy,cw,ih*2);
      g.setColor(Color.black);
      g.drawOval(cx,cy,cw,ih*2);
     }

    g.setColor(Color.yellow);
    g.fillOval(0,d.height-fh,d.width,fh);
    try
     {
      Thread.sleep(100);
     } catch (InterruptedException e) 
     {
     }
    g.setColor(Color.red);
    g.fillOval(0,d.height-fh,d.width,fh);
    try
     {
      Thread.sleep(300);
     } catch (InterruptedException e) 
     {
     }
    n_dropped_jobs++;
    hisw = d.width/bb;
    debris_histo = new int[hisw];
    double sigma = (double)hisw/6;
    double center = (double)hisw/2;
    double gausc = 1/(Math.sqrt(2*Math.PI) * sigma);
    int i;
    double tossx,tossy;
    boolean found;
    for (i=0; i< n_dropped_jobs*3; i++)
     {
      found = false;
      while (!found)
       {
        tossx = Math.random()*(double)hisw;
        tossy = Math.random()*gausc;
        if (tossy < (gausc * Math.exp(-1*Math.pow(tossx-center,2)/2/sigma/sigma)))
         {
          debris_histo[(int)tossx]++;
          break;
         }
       }
     }
    draw_processor(g,d);
    g.clearRect(0,d.height-fh-bb-1,d.width,fh+bb+1);
    draw_debris(g,d);
   }

  void draw_debris(Graphics g, Dimension d)
   {
    if (n_dropped_jobs <= 0) return;
    int i,j,x,y,side;
    double half_side = (double)bb/2;
    for (i=0; i<hisw; i++)
     {
      y = d.height-1;
      for (j=0; j<debris_histo[i]; j++)
       {
        side = (int)(Math.random() * half_side + half_side);
        y-=side;
        if (y < (d.height-fh-bb-1)) break;
        x = i*bb + (bb-side/2);
        g.setColor(qcolor);
        g.fillOval(x,y,side,side);
        g.setColor(Color.black);
        g.drawOval(x,y,side,side);
       }
     }
   }

  void draw_processor(Graphics g, Dimension d)
   {
    int am = d.width/2;

    if (busy) g.setColor(Color.blue);
    else      g.setColor(Color.green);
    if (n_dropped_jobs <= 0) g.clearRect(0,d.height-fh-bb-1,d.width,fh+bb+1);

    g.fillRect(bb,by,bw,bh);
    g.setColor(Color.yellow);
    g.setFont(new Font("dialog", Font.BOLD, bw*18/120));
    String text;
    if (busy) text = new String("Process");
    else      text = new String("Idle");
    FontMetrics cur_fontm = g.getFontMetrics();
    int text_width = cur_fontm.stringWidth(text);
    int text_height = cur_fontm.getHeight();
    g.drawString(text,am-text_width/2,by+bh/2+text_height/2);
   }

  public void paint(Graphics g) 
   {
    Dimension d=getSize();
    int cy;

    init_geometry(d);

    if (n_items>0)
     {
      for (cy=sy; cy>(sy-ih*n_items); cy--)
       {
        if (((cy-sy)%ih)!=0) g.setColor(qcolor);
        else                       g.setColor(Color.black);
  
        g.drawArc(cx,cy,cw,ih*2,180,180);
       } 
      // Draw vertical lines.
      g.setColor(Color.black);
      g.drawLine(cx,sy+ih,cx,sy+ih-ih*n_items);
      g.drawLine(cx+cw,sy+ih,cx+cw,sy+ih-ih*n_items);
  
      // Draw top ellipse.
      g.setColor(qcolor);
      g.fillOval(cx,cy,cw,ih*2);
      g.setColor(Color.black);
      g.drawOval(cx,cy,cw,ih*2);
     }

    n_drawn_items = n_items;
    old_dim = d;

    // Draw box at the bottom and arrow
    int am = d.width/2;

    Polygon arrow1 = new Polygon();
    arrow1.addPoint(am-bb, d.height-2*fh);
    arrow1.addPoint(am+bb, d.height-2*fh);
    arrow1.addPoint(am+bb, by-bb);
    arrow1.addPoint(am+bb+bb, by-bb);
    arrow1.addPoint(am, by);
    arrow1.addPoint(am-bb-bb, by-bb);
    arrow1.addPoint(am-bb, by-bb);
    arrow1.addPoint(am-bb, d.height-2*fh);
    g.setColor(Color.red);
    g.fillPolygon(arrow1);

    Polygon arrow2 = new Polygon();
    arrow2.addPoint(d.width-10*bb,d.height-2*fh+8*bb);
    arrow2.addPoint(d.width-9*bb,d.height-2*fh+9*bb);
    arrow2.addPoint(d.width-bb,d.height-2*fh+bb);
    arrow2.addPoint(d.width,d.height-2*fh+2*bb);
    arrow2.addPoint(d.width,d.height-2*fh-bb);
    arrow2.addPoint(d.width-3*bb,d.height-2*fh-bb);
    arrow2.addPoint(d.width-2*bb,d.height-2*fh);
    arrow2.addPoint(d.width-10*bb,d.height-2*fh+8*bb);
    g.fillPolygon(arrow2);

    draw_processor(g,d);
    draw_debris(g,d);
   }

  public void update (Graphics g) 
   {
    Dimension d=getSize();
    if (old_dim != null)
     {
      if ((d.width != old_dim.width) || (d.height != old_dim.height))
       {
        g.clearRect(0,0,d.width,d.height);
        paint(g);
        return;
       }
     }
    else return;

    draw_processor(g,d);

    if (n_drawn_items == n_items) return;

    int cy;

    if (n_items > n_drawn_items)
     {
      for (cy=sy-ih*n_drawn_items; cy>(sy-ih*n_items); cy--)
       {
        if (((cy-sy)%ih)!=0) g.setColor(qcolor);
        else                 g.setColor(Color.black);
  
        g.drawArc(cx,cy,cw,ih*2,180,180);
       } 
      // Draw vertical lines.
      g.setColor(Color.black);
      g.drawLine(cx,sy+ih,cx,sy+ih-ih*n_items);
      g.drawLine(cx+cw,sy+ih,cx+cw,sy+ih-ih*n_items);
  
      // Draw top ellipse.
      g.setColor(qcolor);
      g.fillOval(cx,cy,cw,ih*2);
      g.setColor(Color.black);
      g.drawOval(cx,cy,cw,ih*2);
     }
    else
     {
      cy = sy-ih*n_items;
      g.clearRect(cx,sy-ih*n_drawn_items,cw+1,(n_drawn_items-n_items)*ih+ih);
  
      // Draw top ellipse.
      if (n_items > 0)
       {
        g.setColor(qcolor);
        g.fillOval(cx,cy,cw,ih*2);
        g.setColor(Color.black);
        g.drawOval(cx,cy,cw,ih*2);
       }
      else
       {
        g.clearRect(cx,sy,cw+1,ih*2+1);
       }
     }

    n_drawn_items = n_items;
    old_dim = d;
   }
 }


class ClockThread extends Thread 
 {
  int count;
  int in_total, out_total;

  StickyQueue controller=null;
  ClockThread(StickyQueue c) 
   {
    controller = c;
    count = 0;
    in_total = out_total = 0;
   }

  public void run() 
   {
    for (;;)
     {
      controller.stime++;
      controller.stimel.setText(new String("Time: "+controller.stime));

      // Determine how many new jobs we want to generate

      double toss = Math.random();
      double poissum = 0;
      int n = 0;
      int nfact = 1;

      for (n=0; n<(5*(int)(controller.in_rate+1)); n++)
       {
        if (n>0) nfact *= n;
        poissum += Math.exp(-1*controller.in_rate) * 
                   Math.pow(controller.in_rate,(double)n) / (double)nfact; 
        if (toss < poissum)
         {
          if (n > 0 && controller.n_queues > 0)
           {
            controller.queuestack[0].add(n);
            in_total += n;
           }
          break;
         }
       }
      
      double in_rate = (double)in_total / (double)controller.stime;

      /* Now check and update all the queues */
      int n_busy_queues = 0;
      int q_total = 0;
      double cload;
      cload = 0;
      for (n=0; n<controller.n_queues; n++)
       {
        if (controller.queuestack[n].busy) 
         {
          n_busy_queues++;
         }
        q_total += controller.queuestack[n].n_items;
        if (controller.queuestack[n].n_items > 1)
         cload += controller.q_stickiness[n] * 
                  (controller.queuestack[n].n_items -1);
       }
      double fraction_available = (1-cload)/(double)n_busy_queues;
      double sticky_fraction;
      if (cload < 1) sticky_fraction = cload;
      else           sticky_fraction = 1;
      if (n_busy_queues == 0) sticky_fraction = 1;

      for (n=0; n<controller.n_queues; n++)
       {
        if (!controller.queuestack[n].busy)
         {
          if (controller.queuestack[n].n_items > 0)
           {
            controller.queuestack[n].busy = true;
            controller.queuestack[n].repaint();
            controller.queuestack[n].ticks_left = controller.q_serviceticks[n];
            controller.queuestack[n].start_time = controller.stime;
           }
         }
        else
         {
          // Did this job time out ?
          if ( (double)(controller.stime - controller.queuestack[n].start_time) >
               controller.q_timeout[n] ) 
           {
            controller.queuestack[n].dec(); 
            controller.queuestack[n].dropjob();
           }
          else
           {
            controller.queuestack[n].ticks_left -= fraction_available;
            if (controller.queuestack[n].ticks_left <= 0)
             {
              controller.queuestack[n].busy = false;
              controller.queuestack[n].dec(); 
              if ((n+1) < controller.n_queues)
               {
                controller.queuestack[n+1].inc();
               }
              else
               {
                out_total++;
               }
             }
           }
         }
       }
      controller.loadg.setSecondaryFraction(sticky_fraction);
      if (n_busy_queues > 0) 
       {
        controller.loadg.set(100);
       }
      else controller.loadg.set((int)(cload*(double)100));
      
      int success_rate=0;
      if (in_total > 0 && out_total > 0)
        success_rate = (int)((double)(out_total+q_total)/(double)in_total * 100);
      if (out_total > 0)
       {
        double out_rate = (double)out_total / (double)controller.stime;
        String text = new Double(out_rate).toString();
        if (text.length() > 10) controller.soutrl.setText(text.substring(0,10));
        else                    controller.soutrl.setText(text);
       }
      controller.success.set(success_rate);

      try
       {
        sleep(400);
       } catch (InterruptedException e) {}
     }
   }
 }

class ParameterDispatch implements TextListener
 {
  StickyQueue mom_applet;
  TextField   mom_timeout;
  TextField   mom_stickiness;
  TextField   mom_serviceticks;
  int mom_queue_index;

  public ParameterDispatch(StickyQueue mom, int n, 
                           TextField timeout, 
                           TextField stickiness, 
                           TextField serviceticks)
   {
    mom_applet = mom;
    mom_timeout = timeout;
    mom_stickiness = stickiness;
    mom_serviceticks = serviceticks;
    mom_queue_index = n;
   }

  public ParameterDispatch(StickyQueue mom) 
   {
    mom_applet = mom;
    mom_timeout = null;
    mom_stickiness = null;
    mom_serviceticks = null;
    mom_queue_index = -1;
   }

  public void textValueChanged(TextEvent e)
   {
    String temp;
    if (mom_queue_index < 0)
     {
      if (e.getSource() == mom_applet.in_ratef)
       {
        temp = mom_applet.in_ratef.getText();
        if (temp != null)
         {
          try
           {
            mom_applet.in_rate = Double.valueOf(temp).doubleValue();
           } catch (NumberFormatException ex) {  
           }
         }
       }
      else return;
     }

    if (e.getSource() == mom_timeout) 
     {
      temp = mom_timeout.getText();
      if (temp != null)
       {
        try
         {
          mom_applet.q_timeout[mom_queue_index] = Double.valueOf(temp).doubleValue();
         } catch (NumberFormatException ex) {  
         }
       }
     }
    else if (e.getSource() == mom_stickiness) 
     {
      temp = mom_stickiness.getText();
      if (temp != null)
       {
        try
         {
          mom_applet.q_stickiness[mom_queue_index] = Double.valueOf(temp).doubleValue();
         } catch (NumberFormatException ex) {  
         }
       }
     }
    else if (e.getSource() == mom_serviceticks) 
     {
      temp = mom_serviceticks.getText();
      if (temp != null)
       {
        try
         {
          mom_applet.q_serviceticks[mom_queue_index] = Double.valueOf(temp).doubleValue();
         } catch (NumberFormatException ex) {  
         }
       }
     }
   }
 }

